/*
 * linux/arch/arm/drivers/char/avr.c
 *
 * Copyright 2001 Applied Data Systems
 *
 */

#include <linux/config.h>

#include <linux/sched.h>
#include <linux/interrupt.h>
#include <linux/kbd_ll.h>
#include <linux/init.h>
#include <linux/delay.h>
#include <linux/kbd_kern.h>

#include <asm/irq.h>
#include <asm/hardware.h>
#include <linux/tqueue.h>
#include <linux/proc_fs.h>
#include <linux/pm.h>
#include "avrprg.h"

#define	DEBUG(args...)	printk(args)

// #define	MAX_LOOP		30
#define	MAX_LOOP		100
static int	flash_size;
static int	flash_family;
static int	flash_part;

typedef	struct	t_AVR_CMD {
	unchar	byte1;
	unchar	byte2; 
	unchar	byte3;
	unchar	byte4; 
}	AVR_CMD;

#if defined(CONFIG_ARCH_ADSAGX)

static void send_SSP_msg(unchar *pBuf, int num)
{
	unchar	tmp;
	int	i;

	for (i=0;i<num;i++) {
		ADSAGX_CPLD_SPI_TX = pBuf[i];

		ADSAGX_CPLD_SPI_EN = ADSAGX_CPLD_SPI_ENABLE;

		udelay(1);
		// wait for transmission to complete
		while ((ADSAGX_CPLD_SPI_EN & ADSAGX_CPLD_SPI_ENABLE) != 0);

		// Throw away SSP echo
		tmp = ADSAGX_CPLD_SPI_RX;
	}
}


static void send_SSP_msg_echo(unchar *pBuf, unchar *echo, int num)
{
	int	i;

	for (i=0;i<num;i++) {
		ADSAGX_CPLD_SPI_TX = pBuf[i];

		ADSAGX_CPLD_SPI_EN = ADSAGX_CPLD_SPI_ENABLE;

		udelay(1);
		// wait for transmission to complete
		while ((ADSAGX_CPLD_SPI_EN & ADSAGX_CPLD_SPI_ENABLE) != 0);

		// save SSP echo
		echo[i] = (unchar) ADSAGX_CPLD_SPI_RX;
	}
}


static inline unchar get_SSP_byte(void)
{
	unchar	data;

	// send 0 to trigger a response
	ADSAGX_CPLD_SPI_TX = 0;

	ADSAGX_CPLD_SPI_EN = ADSAGX_CPLD_SPI_ENABLE;

	udelay(1);
	// wait for transmission to complete
	while ((ADSAGX_CPLD_SPI_EN & ADSAGX_CPLD_SPI_ENABLE) != 0);

	// read the data
	data = (unchar) ADSAGX_CPLD_SPI_RX;

	return (data);
}

#define ReadSSPByte get_SSP_byte
#define set_SSP_highspeed()
#define set_SSP_lowspeed()
#define set_SSP_normalspeed()

#elif defined(CONFIG_ARCH_ADSVGX)
#define MAX_SSP_FIFO_CYCLES 20000;

#  define SSPCR0        NSSCR0
#  define SSPCR1	NSSCR1
#  define SSPSR		NSSSR
#  define SSPDR		NSSDR
// FIXME: This is sloppy - JML
#  undef  SSSR_TNF
#  undef  SSSR_RNE
#  undef  SSSR_ROR
#  define SSSR_TNF	0x00000004	// Transmit FIFO Not Full (PXA)
#  define SSSR_RNE	0x00000008	// Receive FIFO Not Empty (PXA)
#  define SSSR_ROR	0x00000080	// Receive FIFO Over-Run (PXA)

static void send_SSP_msg(unchar *pBuf, int num)
{
	ushort	tmp;
	int	i;

	for (i=0;i<num;i++) {
		while ((SSPSR & SSSR_TNF) == 0);
		tmp = pBuf[i];
		SSPDR = tmp;

		// Throw away SSP echo
		while ((SSPSR & SSSR_RNE) == 0);
		tmp = SSPDR;
	}
}


static void send_SSP_msg_echo(unchar *pBuf, unchar *echo, int num)
{
	ushort	tmp;
	int	i;

	for (i=0;i<num;i++) {
		while ((SSPSR & SSSR_TNF) == 0);
		tmp = pBuf[i];
		SSPDR = tmp;

		// Save SSP echo
		while ((SSPSR & SSSR_RNE) == 0);
		echo[i] = SSPDR;
	}
}


static inline unchar get_SSP_byte(void)
{
	unchar	data;

	// send 0 to trigger a response
	SSPDR = 0x00;

	// wait for rcv not empty status
	while ((SSPSR & SSSR_RNE) == 0);

	// read the data
	data = (unchar) SSPDR;

	return (data);
}

static unchar ReadSSPByte(void)
{
	if (SSPSR & SSSR_ROR) {
		printk(KERN_WARNING "%s() : Overrun\n", __FUNCTION__);
		return 0;
	}

	return get_SSP_byte();
}

static void set_SSP_highspeed(void)
{
	SSPCR0 = 0x1007;
	SSPSR = SSSR_ROR;
	SSPCR1 = 0x0000;
	SSPCR0 = 0x1087;

	mdelay(10);
}

static void set_SSP_lowspeed(void)
{
	SSPCR0 = 0x4007;
	SSPSR = SSSR_ROR;
	SSPCR1 = 0x0000;
	SSPCR0 = 0x4087;
	mdelay(10);
}

static void set_SSP_normalspeed(void)
{
	SSPCR0 = 0xA707;
	SSPSR = SSSR_ROR;
	SSPCR1 = 0x0010;
	SSPCR0 = 0xA787;
	mdelay(10);
}

#elif defined(CONFIG_ARCH_ADSBITSYX)

#define MAX_SSP_FIFO_CYCLES 20000;

#  define SSPCR0        SSCR0
#  define SSPCR1	SSCR1
#  define SSPSR		SSSR
#  define SSPDR		SSDR
// FIXME: This is sloppy - JML
#  undef  SSSR_TNF
#  undef  SSSR_RNE
#  undef  SSSR_ROR
#  define SSSR_TNF	0x00000004	// Transmit FIFO Not Full (PXA)
#  define SSSR_RNE	0x00000008	// Receive FIFO Not Empty (PXA)
#  define SSSR_ROR	0x00000080	// Receive FIFO Over-Run (PXA)

static void send_SSP_msg(unchar *pBuf, int num)
{
	ushort	tmp;
	int	i;

	for (i=0;i<num;i++) {
		while ((SSPSR & SSSR_TNF) == 0);
		tmp = pBuf[i];
		SSPDR = tmp;

		// Throw away SSP echo
		while ((SSPSR & SSSR_RNE) == 0);
		tmp = SSPDR;
	}
}


static void send_SSP_msg_echo(unchar *pBuf, unchar *echo, int num)
{
	ushort	tmp;
	int	i;

	for (i=0;i<num;i++) {
		while ((SSPSR & SSSR_TNF) == 0);
		tmp = pBuf[i];
		SSPDR = tmp;

		// Save SSP echo
		while ((SSPSR & SSSR_RNE) == 0);
		echo[i] = SSPDR;
	}
}


static inline unchar get_SSP_byte(void)
{
	unchar	data;

	// send 0 to trigger a response
	SSPDR = 0x00;

	// wait for rcv not empty status
	while ((SSPSR & SSSR_RNE) == 0);

	// read the data
	data = (unchar) SSPDR;

	return (data);
}

static unchar ReadSSPByte(void)
{
	if (SSPSR & SSSR_ROR) {
		printk(KERN_WARNING "%s() : Overrun\n", __FUNCTION__);
		return 0;
	}

	return get_SSP_byte();
}

static void set_SSP_highspeed(void)
{
	SSPCR0 = 0x1007;
	SSPSR = SSSR_ROR;
	SSPCR1 = 0x0000;
	SSPCR0 = 0x1087;

	mdelay(10);
}

static void set_SSP_lowspeed(void)
{
	SSPCR0 = 0x4007;
	SSPSR = SSSR_ROR;
	SSPCR1 = 0x0000;
	SSPCR0 = 0x4087;
	mdelay(10);
}

static void set_SSP_normalspeed(void)
{
	SSPCR0 = 0xA707;
	SSPSR = SSSR_ROR;
	SSPCR1 = 0x0010;
	SSPCR0 = 0xA787;
	mdelay(10);
}

#endif

static ulong read_SSP_response(int num)
{
	int		i;
	ulong	ret;
	
	// discard leading 0x00 and command echo 0 (command group value)
	ReadSSPByte();
	ReadSSPByte();
	// discard command echo 1 (command code value)
	ReadSSPByte();

	// data from SMARTIO
	// It assumes LSB first.
	// NOTE:Some command uses MSB first order
	ret = 0;
	for (i=0;i<num;i++) {
		ret |= ReadSSPByte() << (8*i);
	}

	return ret;
}

static AVR_CMD ProgramEnable = { 0xAC, 0x53, 0x00, 0x00 };
static AVR_CMD ReadDevCode   = { 0x30, 0x00, 0x00, 0x00 };
static AVR_CMD ReadPrgFlash  = { 0x20, 0x00, 0x00, 0x00 };
static AVR_CMD WritePrgFlash = { 0x40, 0x00, 0x00, 0x00 };
static AVR_CMD ChipErase     = { 0xAC, 0x80, 0x00, 0x00 };

static int send_program_msg(unchar *pBuf, unchar *echo)
{
	send_SSP_msg_echo(pBuf, echo, 4);

	// printk("sent: %#x %#x, got: %#x %#x (should be the same!)\n", pBuf[0], pBuf[1], echo[1], echo[2]);
	if ((pBuf[0] == echo[1]) && (pBuf[1] == echo[2])) {
		return 1;
	}
	else {
		return 0;
	}
}

#if defined(CONFIG_ARCH_ADSAGX) || defined(CONFIG_ARCH_ADSVGX)
#  define ADS_RESETAVR			27
#endif

static void avr_set_reset_low(void)
{
	// Reset SMARTIO
#ifdef CONFIG_ARCH_ADSBITSYX
	ADSBITSYX_CPLD_SUPPC2 |= ADSBITSYX_SUPPC2_RESETAVR;
	mdelay(300);
	ADSBITSYX_CPLD_SUPPC2 &= ~ADSBITSYX_SUPPC2_RESETAVR;
#elif defined(CONFIG_ARCH_ADSAGX) || defined(CONFIG_ARCH_ADSVGX)
	set_GPIO_mode( (ADS_RESETAVR | GPIO_OUT) );
	GPDR(ADS_RESETAVR) |= GPIO_bit(ADS_RESETAVR);
	GPSR(ADS_RESETAVR) |= GPIO_bit(ADS_RESETAVR);
	mdelay(300);
	GPCR(ADS_RESETAVR) |= GPIO_bit(ADS_RESETAVR);
#endif
	mdelay(300);
}

static void avr_reset(void)
{
	// Reset SMARTIO
#ifdef CONFIG_ARCH_ADSBITSYX
	ADSBITSYX_CPLD_SUPPC2 &= ~ADSBITSYX_SUPPC2_RESETAVR;
	mdelay(300);
	ADSBITSYX_CPLD_SUPPC2 |= ADSBITSYX_SUPPC2_RESETAVR;
#elif defined(CONFIG_ARCH_ADSAGX) || defined(CONFIG_ARCH_ADSVGX)
	set_GPIO_mode( (ADS_RESETAVR | GPIO_OUT) );
	GPDR(ADS_RESETAVR) |= GPIO_bit(ADS_RESETAVR);
	GPCR(ADS_RESETAVR) |= GPIO_bit(ADS_RESETAVR);
	mdelay(300);
	GPSR(ADS_RESETAVR) |= GPIO_bit(ADS_RESETAVR);
#endif
	mdelay(300);
}

static int avr_program_enable(void)
{
	unchar	echo[4];
	int		retry;

	avr_reset();

	avr_set_reset_low();

	retry = 0;
	while (!send_program_msg((unchar *)&ProgramEnable, echo)) {
		mdelay(20);
		if (retry++ > 32) break;
	}
	if (retry > 32) 
		return 0;
	else 
		return 1;
}

void avr_program_disable(void)
{
	avr_reset();
}

static int avr_read_devcode(unchar *vendor, unchar *family, unchar *part)
{
	unchar	echo[4];

	ReadDevCode.byte3 = 0x00;
	if (send_program_msg((unchar *)&ReadDevCode, echo)) {
		if (echo[3] && (echo[3] != 0x1E)) {
			printk("AVR Vendor Error: ret = 0x%02x\n", echo[3]);
			return 0;
		}
		*vendor = echo[3];
	}
	else {
	        printk("send_program_msg 1 failed!\n");
		return 0;
	}

	ReadDevCode.byte3 = 0x01;
	if (send_program_msg((unchar *)&ReadDevCode, echo)) {
		if (echo[3] == 0x90) 
			flash_size = 1024;		// 1Kbyte
		else if (echo[3] == 0x91)
			flash_size = 2048;		// 2Kbyte
		else if (echo[3] == 0x92)
			flash_size = 4096;		// 4Kbyte
		else if (echo[3] == 0x93)
			flash_size = 8192;		// 8Kbyte
		else if (echo[3] == 0xFF) {
			printk("AVR Device Code Erased: ret = 0x%02x\n", echo[3]);
			return 0;
		}
		else if (echo[3] != 0x01) {
			printk("AVR Flash size Error: ret = 0x%02x\n", echo[3]);
			return 0;
		}
		*family = flash_family = echo[3];
	}
	else {
	        printk("send_program_msg 2 failed!\n");
		return 0;
	}

	ReadDevCode.byte3 = 0x02;
	if (send_program_msg((unchar *)&ReadDevCode, echo)) {
		*part = flash_part = echo[3];
		if ((*vendor == 0x00) && (*family == 0x01) && (*part == 0x02))
			printk("AVR is locked\n");
	}
	else {
	        printk("send_program_msg 3 failed!\n");
		return 0;
	}

	return 1;
}

void set_SSP_init(int noreset)
{
#if defined(CONFIG_ARCH_ADSBITSYX)
	// Configure the Alternate Function and 
	// Direction Registers for the SSP Pins
	set_GPIO_mode(GPIO23_SCLK_md);
	set_GPIO_mode(GPIO24_SFRM_MD);
	set_GPIO_mode(GPIO25_STXD_MD);
	set_GPIO_mode(GPIO26_SRXD_MD);

	// Enable SSP Unit Clock
	CCCR |= CKEN3_SSP;

	// Configure the SSP Serial Port
	SSPCR0  = SSCR0_Motorola | SSCR0_SerClkDiv(336) | SSCR0_DataSize(8);
	SSPSR   = SSSR_ROR;
	SSPCR1 |= SSCR1_SP | 0x08;
	SSPCR0 |= SSCR0_SSE;

	// Reset SMARTIO
	if (ADSBITSYX_REV < 2) {
		printk("Error: OLD REV Bitsy board, can't update AVR!\n");
#if 0
		set_GPIO_mode( (ADSBITSYX_RESETAVR | GPIO_OUT) );
		GPDR(ADSBITSYX_RESETAVR) |= GPIO_bit(ADSBITSYX_RESETAVR);
		GPCR(ADSBITSYX_RESETAVR) |= GPIO_bit(ADSBITSYX_RESETAVR);
		mdelay(300);
		GPSR(ADSBITSYX_RESETAVR) |= GPIO_bit(ADSBITSYX_RESETAVR);
#endif
	} else {
		printk("Reseting BitsyX\n");
		ADSBITSYX_CPLD_SUPPC2 &= ~ADSBITSYX_SUPPC2_RESETAVR;
		mdelay(300);
		ADSBITSYX_CPLD_SUPPC2 |= ADSBITSYX_SUPPC2_RESETAVR;
	}

#elif defined(CONFIG_ARCH_ADSAGX)

	if (noreset) {
		printk("not resetting AVR\n");
		return;
	}

	// Reset SMARTIO
	printk("Reseting AGX\n");
	set_GPIO_mode( (ADS_RESETAVR | GPIO_OUT) );
	GPDR(ADS_RESETAVR) |= GPIO_bit(ADS_RESETAVR);
	GPCR(ADS_RESETAVR) |= GPIO_bit(ADS_RESETAVR);
	mdelay(300);
	GPSR(ADS_RESETAVR) |= GPIO_bit(ADS_RESETAVR);
#elif defined(CONFIG_ARCH_ADSVGX)

	// Enable SSP Controller Clock
	set_GPIO_mode(GPIO81_NSSPCLK_OUT_MD);
	set_GPIO_mode(GPIO82_NSSPSFRM_OUT_MD);
	set_GPIO_mode(GPIO83_NSSPTXD_OUT_MD);
	set_GPIO_mode(GPIO84_NSSPRXD_IN_MD);

	// Enable SSP Unit Clock
	CKEN |= CKEN9_NSSP;

	// Configure the SSP Serial Port
	// National Microwire - 450k baud - 16 bit
	NSSCR0  = SSCR0_Motorola | NSSCR0_SerClkDiv(336) | NSSCR0_DataSize(8);
	NSSSR   = NSSSR_ROR;
	NSSCR1 &= ~(SSCR1_RIE | NSSCR1_TIE | NSSCR1_LBM | NSSCR1_MWDS);
	NSSCR1 |= (0x77<<7);
	NSSCR0 |= NSSCR0_SSE;

	// Reset SMARTIO
	printk("Reseting VGX\n");
	set_GPIO_mode( (ADS_RESETAVR | GPIO_OUT) );
	GPDR(ADS_RESETAVR) |= GPIO_bit(ADS_RESETAVR);
	GPCR(ADS_RESETAVR) |= GPIO_bit(ADS_RESETAVR);
	mdelay(300);
	GPSR(ADS_RESETAVR) |= GPIO_bit(ADS_RESETAVR);
#endif
	mdelay(300);
}

static	AVR_CMD READ_DEVVER_CMD  = { 0x82, 0x05, 0x00, 0x00 };
static	AVR_CMD READ_DEVTYPE_CMD = { 0x82, 0x06, 0x00, 0x00 };
static	AVR_CMD READ_FWLEVEL_CMD = { 0x82, 0x07, 0x00, 0x00 };

static ushort read_sio_version(struct sio_ver *ptr)
{
	ushort			ret;

	// Read Device Version 
	send_SSP_msg((unchar *) &READ_DEVVER_CMD, 2);
	ret = read_SSP_response(1);
	ptr->DevVer = (uint)ret;
	// Read Device Type
	send_SSP_msg((unchar *) &READ_DEVTYPE_CMD, 2);
	ret = read_SSP_response(2);
	ptr->DevType = (uint)ret;
	// Read Firmware Level
	send_SSP_msg((unchar *) &READ_FWLEVEL_CMD, 2);
	ret = read_SSP_response(2);
	// swap MSB & LSB
	ret = ((ret & 0xFF) << 8) | ((ret & 0xFF00) >> 8);
	ptr->FwLevel = (uint)ret;

	return 0;
}

static int avr_open(struct inode *inode, struct file *file)
{
	unchar	vendor, family, part;

	set_SSP_highspeed();
	if (!avr_program_enable()) {
                printk("Unable to enable programming.\n");
		return -EIO;
	}

	if (!avr_read_devcode(&vendor, &family, &part)) {
                printk("Unable to read device code.\n");
		return -EIO;
	}

	avr_program_disable();
	set_SSP_normalspeed();

	return 0;
}


static int avr_release(struct inode *inode, struct file *file)
{
	avr_program_disable();

	set_SSP_normalspeed();

	return 0;
}

static ssize_t avr_read(struct file *file, char *buf, size_t count, loff_t *ppos)
{
	struct inode *inode = file->f_dentry->d_inode;
	unsigned int minor = MINOR(inode->i_rdev);
	unchar	*data = (unchar *)buf;
	unchar	upper, lower;
	unchar	echo[4];
	int		i, retry;

	if (minor != 0) return -ENXIO;

	if (count > flash_size) {
		printk("Requested size is greater than AVR Flash Size\n");
		return -EINVAL;
	}

	set_SSP_highspeed();

	if (!avr_program_enable())
		return -EIO;

	upper = lower = 0;
	for (i=0;i<count/2;i++) {
		ReadPrgFlash.byte1 = 0x20;			// Low Byte
		ReadPrgFlash.byte2 = upper;
		ReadPrgFlash.byte3 = lower;
		retry = 0;
		while (!send_program_msg((unchar *)&ReadPrgFlash, echo)) {
			if (retry++ > MAX_LOOP) {
				avr_program_disable();
				printk("Device is busy..(1)\n");
				return -EBUSY;
			}
		}
		*data++ = echo[3];

		ReadPrgFlash.byte1 = 0x28;			// High Byte
		retry = 0;
		while (!send_program_msg((unchar *)&ReadPrgFlash, echo)) {
			if (retry++ > MAX_LOOP) {
				avr_program_disable();
				printk("Device is busy..(2)\n");
				return -EBUSY;
			}
		}
		*data++ = echo[3];

		if (lower == 0xFF) upper++;
		lower ++;
	}

	if (count%2) {
		ReadPrgFlash.byte1 = 0x20;			// Low Byte
		ReadPrgFlash.byte2 = upper;
		ReadPrgFlash.byte3 = lower;
		retry = 0;
		while(!send_program_msg((unchar *)&ReadPrgFlash, echo)) {
			if (retry++ > MAX_LOOP) {
				avr_program_disable();
				printk("Device is busy..(3)\n");
				return -EBUSY;
			}
		}
		*data++ = echo[3];
	}
	avr_program_disable();

	return count;
}

static ssize_t avr_write_old(struct file *file, const char *buf, size_t count, loff_t *ppos)
{
	struct inode *inode = file->f_dentry->d_inode;
	unsigned int minor = MINOR(inode->i_rdev);
	unchar	echo[4];
	unchar	upper, lower;
	unchar	*data = (unchar *)buf;
	int		i, retry;

	if (minor != 0) return -ENXIO;

	if (count > flash_size) {
		printk("Requested size is greater than AVR Flash Size\n");
		return -EINVAL;
	}
	if (!avr_program_enable())
		return -EIO;

	set_SSP_lowspeed();

	upper = lower = 0;
	for (i=0;i<count/2;i++) {
		WritePrgFlash.byte1 = 0x40;			// Low Byte
		WritePrgFlash.byte2 = upper;
		WritePrgFlash.byte3 = lower;
		WritePrgFlash.byte4 = *data;
		while (!send_program_msg((unchar *)&WritePrgFlash, echo));
		mdelay(5);
		ReadPrgFlash.byte1 = 0x20;
		ReadPrgFlash.byte2 = upper;
		ReadPrgFlash.byte3 = lower;
		for (retry=0;retry<MAX_LOOP;retry++) {
			if (*data == 0xFF)  {
				mdelay(10);
				break;	
			}

			send_program_msg((unchar *)&ReadPrgFlash, echo);
			if (*data == echo[3])
				break;
			mdelay(5);
		}
		if (retry == MAX_LOOP) {
			avr_program_disable();
			printk("Device is busy..(4) %d\n", i);
			return -EBUSY;
		}
		else
			data++;

		WritePrgFlash.byte1 = 0x48;			// High Byte
		WritePrgFlash.byte2 = upper;
		WritePrgFlash.byte3 = lower;
		WritePrgFlash.byte4 = *data;
		while (!send_program_msg((unchar *)&WritePrgFlash, echo));
		mdelay(5);
		ReadPrgFlash.byte1 = 0x28;
		ReadPrgFlash.byte2 = upper;
		ReadPrgFlash.byte3 = lower;
		for (retry=0;retry<MAX_LOOP;retry++) {
			if (*data == 0xFF)  {
				mdelay(10);
				break;	
			}

			send_program_msg((unchar *)&ReadPrgFlash, echo);
			if (*data == echo[3])
				break;
			mdelay(5);
		}
		if (retry == MAX_LOOP) {	
			avr_program_disable();
			printk("Device is busy..(5) %d\n", i);
			return -EBUSY;
		}
		else
			data++;

		if (lower == 0xFF) upper++;
		lower ++;
	}

	if (count%2) {
		WritePrgFlash.byte1 = 0x40;			// Low Byte
		WritePrgFlash.byte2 = upper;
		WritePrgFlash.byte3 = lower;
		WritePrgFlash.byte4 = *data;
		while (!send_program_msg((unchar *)&WritePrgFlash, echo));
		mdelay(5);
		ReadPrgFlash.byte1 = 0x20;
		ReadPrgFlash.byte2 = upper;
		ReadPrgFlash.byte3 = lower;
		for (retry=0;retry<MAX_LOOP;retry++) {
			if (*data == 0xFF)  {
				mdelay(10);
				break;	
			}
			send_program_msg((unchar *)&ReadPrgFlash, echo);
			if (*data == echo[3])
				break;
			mdelay(5);
		}
		if (retry == MAX_LOOP) {
			avr_program_disable();
			printk("Device is busy..(6)\n");
			return -EBUSY;
		}
	}

	set_SSP_highspeed();
	avr_program_disable();

	return count;
}

static ssize_t avr_write_new(struct file *file, const char *buf, size_t count, loff_t *ppos)
{
	struct inode *inode = file->f_dentry->d_inode;
	unsigned int minor = MINOR(inode->i_rdev);
	unchar	echo[4];
	unchar	upper, lower;
	unchar	*data = (unchar *)buf;
	int	i;
	int     page, page_count;

	if (minor != 0) return -ENXIO;

	if (count > flash_size) {
		printk("Requested size is greater than AVR Flash Size\n");
		return -EINVAL;
	}
	if (!avr_program_enable())
		return -EIO;

	set_SSP_lowspeed();

	upper = lower = 0;

	page_count = count / 64;

	for (page=0; page < page_count; page++) {

	  for (i=0; i<32; i++) {
		WritePrgFlash.byte1 = 0x40;			// Low Byte
		WritePrgFlash.byte2 = 0x00;
		WritePrgFlash.byte3 = i;
		WritePrgFlash.byte4 = *data;
		while (!send_program_msg((unchar *)&WritePrgFlash, echo));

		data++;
		WritePrgFlash.byte1 = 0x48;			// Low Byte
		WritePrgFlash.byte2 = 0x00;
		WritePrgFlash.byte3 = i;
		WritePrgFlash.byte4 = *data;
		while (!send_program_msg((unchar *)&WritePrgFlash, echo));
		data++;
	  }

	  // Write page to flash
	  WritePrgFlash.byte1 = 0x4C;
	  WritePrgFlash.byte2 = ((page >> 3) & 0xF);
	  WritePrgFlash.byte3 = ((page & 0x7) << 5);
	  WritePrgFlash.byte4 = 0x00;
	  send_program_msg((unchar *)&WritePrgFlash, echo);
	  mdelay(5);
	}

	if (count % 64) {

	  for (i=0; i < ((count % 64) / 2); i++) {
		WritePrgFlash.byte1 = 0x40;			// Low Byte
		WritePrgFlash.byte2 = 0x00;
		WritePrgFlash.byte3 = i;
		WritePrgFlash.byte4 = *data;
		while (!send_program_msg((unchar *)&WritePrgFlash, echo));
		data++;

		WritePrgFlash.byte1 = 0x48;			// Low Byte
		WritePrgFlash.byte2 = 0x00;
		WritePrgFlash.byte3 = i;
		WritePrgFlash.byte4 = *data;
		while (!send_program_msg((unchar *)&WritePrgFlash, echo));
		data++;
	  }

	  if (count & 0x1) {
		WritePrgFlash.byte1 = 0x40;			// Low Byte
		WritePrgFlash.byte2 = 0x00;
		WritePrgFlash.byte3 = i;
		WritePrgFlash.byte4 = *data;
		while (!send_program_msg((unchar *)&WritePrgFlash, echo));
		data++;

	  }

	  // Write page to flash
	  page = page_count;
	  WritePrgFlash.byte1 = 0x4C;
	  WritePrgFlash.byte2 = ((page >> 3) & 0xF);
	  WritePrgFlash.byte3 = ((page & 0x7) << 5);
	  WritePrgFlash.byte4 = 0x00;
	  send_program_msg((unchar *)&WritePrgFlash, echo);
	  mdelay(5);
	}

	// high fuse
	mdelay(5);
	WritePrgFlash.byte1 = 0xAC;
	WritePrgFlash.byte2 = 0xA8;
	WritePrgFlash.byte3 = 0x00;
	WritePrgFlash.byte4 = 0x49;
	send_program_msg((unchar *)&WritePrgFlash, echo);

	// fuse
	mdelay(5);
	WritePrgFlash.byte1 = 0xAC;
	WritePrgFlash.byte2 = 0xA0;
	WritePrgFlash.byte3 = 0x00;
	WritePrgFlash.byte4 = 0xEF;
	send_program_msg((unchar *)&WritePrgFlash, echo);

	mdelay(100);

	set_SSP_highspeed();
	avr_program_disable();

	return count;
}

static ssize_t avr_write(struct file *file, const char *buf, size_t count, loff_t *ppos)
{
        int result;
	if (flash_family == 0x93 && flash_part == 0x08) {
                // ATmega new part
	  result = avr_write_new(file, buf, count, ppos);
	}
	else {
	  result = avr_write_old(file, buf, count, ppos);
	}

	return result;
}

static int avr_ioctl(struct inode *inode, struct file *file, unsigned int cmd, unsigned long arg)
{
	unsigned int minor = MINOR(inode->i_rdev);
	unchar	vendor, family, part, echo[4];
	struct avr_ver	*DevVer;

	if (minor != 0) return -ENXIO;

	set_SSP_highspeed();

	switch (cmd) {
		case	READ_AVR_VER:
				set_SSP_normalspeed();
				read_sio_version((struct sio_ver *) arg);
				mdelay(100);				// delay for stable operation
				return sizeof(struct sio_ver);
		case	READ_AVR_DEV:
				set_SSP_highspeed();
				if (!avr_program_enable())
					return -EIO;
				if (!avr_read_devcode(&vendor, &family, &part)) {
					avr_program_disable();
					return -EIO;
				}
				else {
					DevVer = (struct avr_ver *)arg;
					DevVer->Vendor = vendor;
					DevVer->Family = family;
					DevVer->Part = part;
					return 3*sizeof(unchar);
				}
		case	READ_FLASH_SIZE:
				(*(int *)arg) = flash_size;
				return flash_size;
		case	ERASE_CHIP:
				set_SSP_highspeed();
				if (!avr_program_enable())
					return -EIO;
				if (!send_program_msg((unchar *)&ChipErase, echo))
					return -EBUSY;
				mdelay(30);
				avr_program_disable();
				break;
		default:
			return -EINVAL;
	}

	return 1;
}

struct file_operations avr_fops = {
	open:		avr_open,
	release:	avr_release,
	read: 		avr_read,
	write:		avr_write,
	ioctl:		avr_ioctl,
};
