#include <stdio.h>
#include <sys/types.h>
#include <fcntl.h>
#include <sys/ioctl.h>
#include <sys/stat.h>
#include "avrprg.h"

unsigned char buf[0x2000];			// 8Kbyte Max
unsigned char rbuf[0x2000];

struct sio_ver	a;
struct avr_ver	b;

struct stat filestat;

main(int argc, char *argv[])
{
	unsigned char *p;
	int	fd, file;
	int	size, count, i, ret;

	filestat.st_size = 0;

	fprintf(stdout, "ADS AVR Programmer for Linux. Ver 1.0 (Build:%s)\n", 
			__DATE__);

	if (argc == 2) {
		if ((file = open(argv[1], O_RDONLY)) == -1) {
			fprintf(stderr, "Can not open %s\n", argv[1]);
			exit(0);
		}

		fstat(file, &filestat);
		if (filestat.st_size < 0x2000)
			read(file, rbuf, filestat.st_size);
		else {
			fprintf(stderr, "File is too big to read\n");
			exit(0);
		}
		close(file);
	}

	if ((fd = open("/dev/avr", O_RDWR)) == -1) {
		fprintf(stderr, "Can not Open /dev/avr device\n");
		exit(0);
	}

	ioctl(fd, LED_GREEN_OFF, NULL);
	ioctl(fd, LED_RED_OFF, NULL);
	ioctl(fd, LED_AMBER_ON, NULL);
	ioctl(fd, READ_AVR_VER, &a);
	fprintf(stdout, "SmartIO has DevVer: %c [0x%x], DevType: 0x%x, Firmware: 0x%x\n",
			a.DevVer, a.DevVer, a.DevType, a.FwLevel);

	ioctl(fd, READ_AVR_DEV, &b);
	if (b.Vendor == 0x1E) {
		if ((b.Family == 0x92) && (b.Part == 0x02))
			fprintf(stdout, "AVR is AT90S4434 with 4Kbyte Flash\n");
		else if ((b.Family == 0x93) && (b.Part == 0x03))
			fprintf(stdout, "AVR is AT90S8535 with 8Kbyte Flash\n");
		else 
			fprintf(stdout, "AVR Family is 0x%x, Part is 0x%x\n",
					b.Family, b.Part);
	}
	else
		fprintf(stdout, "Unkown AVR vendor code : 0x%x (0x%x, 0x%x)\n", 
				b.Vendor, b.Family, b.Part);

	if (filestat.st_size) {
		ioctl(fd, READ_FLASH_SIZE, &size);

		fprintf(stdout, "Erasing Flash/EEPROM of AVR..");
		fflush(stdout);
		if (ioctl(fd, ERASE_CHIP, NULL) == -1) {
			ioctl(fd, LED_AMBER_OFF, NULL);
			ioctl(fd, LED_RED_ON, NULL);
			fprintf(stderr, "AVR is not erased\n");
			close(fd);
			exit(0);
		}
		fprintf(stdout, "Done\n");

		/*
		fprintf(stdout, "Check AVR is Blank\n");
		fprintf(stdout, "\t Read AVR..");
		fflush(stdout);
		if((ret = read(fd, buf, size)) == -1) {
			fprintf(stderr, "Error\n");
			exit(0);
		}
		fprintf(stdout, "Done\n");

		fprintf(stdout, "\t Verify Data..");
		fflush(stdout);
		for (i=0;i<size;i++) {
			if (buf[i] != 0xFF) {
				fprintf(stderr, "AVR is not Erased\n");
				fprintf(stderr, "buf[%d] is not 0xFF but 0x%02x\n", i, buf[i]);
			}
		}
		fprintf(stdout, "OK\n");
		*/

		fprintf(stdout, "Program Data..");
		fflush(stdout);
		count = filestat.st_size / 10;
			if ((ret = write(fd, rbuf, filestat.st_size)) == -1) {
				ioctl(fd, LED_AMBER_OFF, NULL);
				ioctl(fd, LED_RED_ON, NULL);
				fprintf(stderr, "Error\n");
				exit(0);
			}
		fprintf(stdout, "Done\n");

		fprintf(stdout, "Verify AVR\n");
		fprintf(stdout, "\t Read AVR..");
		fflush(stdout);
		if((ret = read(fd, buf, filestat.st_size)) == -1) {
			ioctl(fd, LED_AMBER_OFF, NULL);
			ioctl(fd, LED_RED_ON, NULL);
			fprintf(stderr, "Error\n");
			close(fd);
			exit(0);
		}
		fprintf(stdout, "Done\n");

		fprintf(stdout, "\t Compare Data..");
		for (i=0;i<filestat.st_size;i++) {
			if (rbuf[i] != buf[i]) {
				fprintf(stderr, "Data should be 0x%02x at %d, but 0x%02x\n",
						rbuf[i], i, buf[i]);
				fprintf(stderr, "Error\n");
				ioctl(fd, LED_AMBER_OFF, NULL);
				ioctl(fd, LED_RED_ON, NULL);
				close(fd);
				exit(0);
			}
		}
		fprintf(stdout, "Done\n");
	}
	else
		fprintf(stdout, "File size is zero\n");

	ioctl(fd, LED_AMBER_OFF, NULL);
	ioctl(fd, LED_GREEN_ON, NULL);
	exit(1);
	close(fd);
}
